/*----------------------------------------------------------------------
  File    : xskel.c
  Contents: X11 program skeleton
  Author  : Christian Borgelt
  History : 15.05.1998 file created
            03.06.1998 minor improvements
            06.08.1998 minor improvements
            03.06.1999 basic drawing functionality added
            10.11.1999 mouse drawing added
            12.01.2000 adapted to module dialog
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- extensions --- */
#define MINXEXT       64        /* minimal x-extension of window */
#define MINYEXT       40        /* minimal y-extension of window */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     240        /* initial y-extension of window */

/* --- colors --- */
#define WHITE          0        /* white */
#define BLACK          1        /* black */
#define RED            2        /* (light) red */
#define GREEN          3        /* (light) green */
#define YELLOW         4        /* yellow */
#define BLUE           5        /* (light) blue */
#define PURPLE         6        /* purple */
#define GREY           7        /* grey */
#define GRAY           7        /* grey (Am.) */

/* --- error codes --- */
#define OK             0        /* no error */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_DIALOG     (-7)       /* cannot create dialog box */
#define E_COLOR      (-8)       /* color allocation failed */
#define E_FONT       (-9)       /* font loading failed */
#define E_GRAPH     (-10)       /* cannot create graphics context */
#define E_UNKNOWN   (-11)       /* unkown error */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Function Prototypes
----------------------------------------------------------------------*/
static void mouse    (Widget w, XEvent *e, String *s, Cardinal *c);
static void load     (Widget w, XEvent *e, String *s, Cardinal *c);
static void save     (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit     (Widget w, XEvent *e, String *s, Cardinal *c);
static void dialog   (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw   (Widget w, XEvent *e, String *s, Cardinal *c);
static void about    (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xskel.rsc"            /* fallback resources */

static XtActionsRec actions[] = {
  { "mouse",    mouse   },      /* mouse button pressed */
  { "load",     load     },     /* File    > Load File... */
  { "save",     save     },     /* File    > Save File... */
  { "quit",     quit     },     /* File    > Quit */
  { "dialog",   dialog   },     /* Actions > Dialog */
  { "redraw",   redraw   },     /* Actions > Redraw */
  { "about",    about    },     /* Help    > About XSkel... */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XSkel.Error",
                         "no error",                          0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XSkel.Error",
                         "not enough memory",                 0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XSkel.Warning",
                         "cannot open file:\n%s",             1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XSkel.Warning",
                         "read error on file:\n%s",           1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XSkel.Warning",
                         "write error on file:\n%s",          1 },
  /* E_INIT     -5 */  { "error",   "init",    "XSkel.Error",
                         "initialization failed",             0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "XSkel.Error",
                         "widget creation failed",            0 },
  /* E_DIALOG   -7 */  { "warning", "dialog",  "XSkel.Warning",
                         "cannot create dialog box",          0 },
  /* E_COLOR    -8 */  { "error",   "color",   "XSkel.Error",
                         "color allocation failed",           0 },
  /* E_FONT     -9 */  { "error",   "font",    "XSkel.Error",
                         "font loading failed",               0 },
  /* E_GRAPH   -10 */  { "error",   "gc",      "XSkel.Error",
                         "cannot create graphics context",    0 },
  /* E_UNKNOWN -11 */  { "error", "unknown",   "XSkel.Error",
                         "unknown error",                     0 } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for maze */

/* --- property sheet --- */
static PSHEET  *psh = NULL;     /* property sheet */

/* --- colors --- */
static unsigned long colors[8]; /* basic colors */

/* --- fonts --- */
static XFontStruct *font;       /* font information */

/* --- parameters --- */
struct {                        /* parameter structure */
  char text[80];                /* a text */
  int  number;                  /* a number */
  int  select;                  /* a selection from a list */
  int  flag;                    /* a flag */
  char fname[1024];             /* a filename */
} params = {                    /* initial values for these fields */
  "abc", 123, 0, 0, "noname.dat"
};

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data and */
  va_start(va, code);           /* list of variable arguments */
  for (argc = 0; argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is ``error'' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is ``warning'' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load File...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save File...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_dialog (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Dialog...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  printf((ok) ? "\nbutton: ok\n" : "\nbutton: cancel\n");
  if (!ok) return;              /* if not ok pressed, abort */

  strcpy(params.text,  psh_getval(psh, "text"));
  params.number = atoi(psh_getval(psh, "number"));
  params.select = (int)psh_getval(psh, "select");
  params.flag   = (int)psh_getval(psh, "check");
  strcpy(params.fname, psh_getval(psh, "file"));

  printf("result of dialog:\n");
  printf("text  : %s\n", params.text);
  printf("number: %i\n", params.number);
  printf("select: %i\n", params.select);
  printf("flag  : %i\n", params.flag);
  printf("fname : %s\n", params.fname);
}  /* dcb_dialog() */

/*--------------------------------------------------------------------*/

static void dcb_browse (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Browse...' button callback */
  psh_setval(psh, "file",fname);/* set selected file name */
}  /* dcb_browse() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_dialog (Widget widget, XtPointer client, XtPointer call)
{ dialog(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

static void mcb_browse (Widget widget, XtPointer client, XtPointer call)
{ fselect(w_top, "Browse...", NULL, NULL, dcb_browse); }  

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void mouse (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- mouse button pressed */
  int colid;                    /* color identifier */

  if (*c != 1) return;          /* if no parameter available */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
    XSetFont(display, gc, font->fid);
  }                             /* create a graphics context */
  if      (strcmp(*s, "1") == 0) colid = RED;
  else if (strcmp(*s, "2") == 0) colid = GREEN;
  else if (strcmp(*s, "3") == 0) colid = BLUE;
  else return;                  /* get the color to draw with */
  XSetForeground(display, gc, colors[colid]);
  XDrawArc(display, window, gc, /* draw an arc at the mouse position */
           e->xbutton.x-3, e->xbutton.y-3, 6, 6, 0, 360*64);
  XFillArc(display, window, gc, /* draw a disk at the mouse position */
           e->xbutton.x-3, e->xbutton.y-3, 6, 6, 0, 360*64);
}  /* mouse() */

/*--------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void dialog (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Dialog */
  char buf[32];                 /* buffer for number */

  if (!psh) {                   /* if property sheet not created yet */
    psh = psh_create("dialog", w_top, 1);
    if (!psh) return;           /* create a property sheet */
    if ((psh_addline(psh, "ltext")                  != 0)
    ||  (psh_additem(psh, "text",   PSH_EDIT,    0) != 0)
    ||  (psh_addline(psh, "lnumber")                != 0)
    ||  (psh_additem(psh, "number", PSH_EDIT,    0) != 0)
    ||  (psh_addline(psh, "lselect")                != 0)
    ||  (psh_additem(psh, "select", PSH_MENU,    0) != 0)
    ||  (psh_menuadd(psh, "select", "first")        != 0)
    ||  (psh_menuadd(psh, "select", "second")       != 0)
    ||  (psh_menuadd(psh, "select", "third")        != 0)
    ||  (psh_menuadd(psh, "select", "last")         != 0)
    ||  (psh_additem(psh, "check",  PSH_TOGGLE,  0) != 0)
    ||  (psh_addline(psh, "lfile")                  != 0)
    ||  (psh_additem(psh, "browse", PSH_COMMAND, 0) != 0)
    ||  (psh_addline(psh, NULL)                     != 0)
    ||  (psh_additem(psh, "file",   PSH_EDIT,    0) != 0)) {
      psh_delete(psh); error(E_WIDGET); }
    psh_addcb(psh, "browse", mcb_browse, NULL);
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh, "text", params.text);
  sprintf(buf, "%i", params.number);
  psh_setval(psh, "number", buf);
  psh_setval(psh, "select", (const char*)params.select);
  psh_setval(psh, "check",  (const char*)params.flag);
  psh_setval(psh, "file",   params.fname);
  psh_handle(psh, dcb_dialog);   /* handle dialog */
}  /* dialog() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static int init = 0;          /* initialization flag */
  XRectangle rc;                /* clipping rectangle */
  int        i;                 /* loop variable */
  int        x, y;              /* coordinates */
  char       *msg;              /* a message text */
  int        len;               /* length of message text */

  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
    XSetFont(display, gc, font->fid);
  }                             /* create a graphics context */
  if (e && (e->type == Expose)){/* if called by Xt, */
    rc.x      = e->xexpose.x;   /* build and set */
    rc.y      = e->xexpose.y;   /* clipping rectangle */
    rc.width  = e->xexpose.width;
    rc.height = e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted);
  }

  /* --- some example graphics --- */
  XSetForeground(display, gc, colors[WHITE]);
  XFillRectangle(display, window, gc, 0, 0, INITXEXT, INITYEXT);

  XSetForeground(display, gc, colors[RED]);
  XFillArc(display, window, gc,  40,  30, 21, 21, 0, 360*64);
  XSetForeground(display, gc, colors[BLUE]);
  XFillArc(display, window, gc,  70,  60, 21, 21, 0, 360*64);
  XSetForeground(display, gc, colors[GREEN]);
  XFillArc(display, window, gc, 100,  90, 21, 21, 0, 360*64);
  XSetForeground(display, gc, colors[YELLOW]);
  XFillArc(display, window, gc, 130, 120, 21, 21, 0, 360*64);

  XSetForeground(display, gc, colors[RED]);
  XDrawArc(display, window, gc, 130,  30, 20, 20, 0, 360*64);
  XDrawArc(display, window, gc, 133,  33, 14, 14, 0, 360*64);
  XDrawArc(display, window, gc, 136,  36,  8,  8, 0, 360*64);
  XSetForeground(display, gc, colors[BLUE]);
  XDrawArc(display, window, gc, 160,  60, 20, 20, 0, 360*64);
  XDrawArc(display, window, gc, 163,  63, 14, 14, 0, 360*64);
  XDrawArc(display, window, gc, 166,  66,  8,  8, 0, 360*64);
  XSetForeground(display, gc, colors[GREEN]);
  XDrawArc(display, window, gc, 190,  90, 20, 20, 0, 360*64);
  XDrawArc(display, window, gc, 193,  93, 14, 14, 0, 360*64);
  XDrawArc(display, window, gc, 196,  96,  8,  8, 0, 360*64);
  XSetForeground(display, gc, colors[YELLOW]);
  XDrawArc(display, window, gc, 220, 120, 20, 20, 0, 360*64);
  XDrawArc(display, window, gc, 223, 123, 14, 14, 0, 360*64);
  XDrawArc(display, window, gc, 226, 126,  8,  8, 0, 360*64);

  XSetForeground(display, gc, colors[BLACK]);
  XDrawLine(display, window, gc,  10,  25, 135, 150);
  XDrawLine(display, window, gc,  60,  25, 185, 150);
  XDrawLine(display, window, gc,  10,  25,  60,  25);
  XDrawLine(display, window, gc, 135, 150, 185, 150);

  XSetForeground(display, gc, colors[WHITE]);
  if (!init) { srand((unsigned)time(NULL)); init = 1; }
  XSetForeground(display, gc, colors[GREY]);
  for (i = 0; i < 1000; i++) {
    x = (int)(rand()/(double)RAND_MAX *50);
    y = (int)(rand()/(double)RAND_MAX *50);
    XDrawPoint(display, window, gc, x +250, y +25);
  }

  XSetForeground(display, gc, colors[BLACK]);
  msg = "This is a test.";
  i   = XTextWidth(font, msg, len = strlen(msg));
  XDrawString(display, window, gc, 160 -i/2, 200, msg, len);

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, dcb_nop) != 0)
    error(E_DIALOG);            /* show `about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  int      n;                   /* argument counter */
  Colormap colmap;              /* color map of display */
  XColor   xc[7];               /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XSkel",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",    0, 0,          0) != 0)
  ||  (mn_additem (menu, "load",    0, mcb_load,   0) != 0)
  ||  (mn_additem (menu, "save",    0, mcb_save,   0) != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",    0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions", 0, 0,          0) != 0)
  ||  (mn_additem (menu, "dialog",  0, mcb_dialog, 0) != 0)
  ||  (mn_additem (menu, "redraw",  0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "help",    0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",   0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,      0, 0,          0) != 0))
    error(E_WIDGET);
  mn_resize(menu, INITXEXT);    /* set width of menu bar */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);
  
  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  if ((XAllocNamedColor(display, colmap, "red",    xc+1, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "green",  xc+2, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "yellow", xc+3, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "blue",   xc+4, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "purple", xc+5, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "grey",   xc+6, xc) == 0))
    error(E_COLOR);             /* allocate colors */
  colors[WHITE ] = WhitePixel(display, screen);
  colors[BLACK ] = BlackPixel(display, screen);
  colors[RED   ] = xc[1].pixel; /* enter the color values */
  colors[GREEN ] = xc[2].pixel; /* into the color vector */
  colors[YELLOW] = xc[3].pixel;
  colors[BLUE  ] = xc[4].pixel;
  colors[PURPLE] = xc[5].pixel;
  colors[GREY  ] = xc[6].pixel;

  /* --- load font --- */
  font = XLoadQueryFont(display, "fixed");
  if (!font) error(E_FONT);     /* load fixed font */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top),
    "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return `ok' */
}  /* main() */
